# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Alluka's standard injection client implementation."""

# pyright: reportOverlappingOverload=warning

from __future__ import annotations

__all__: list[str] = ["Client", "inject"]

import asyncio
import enum
import typing
import warnings
import weakref
from collections import abc as collections

import typing_extensions

from . import _context  # pyright: ignore[reportPrivateUsage]
from . import _errors  # pyright: ignore[reportPrivateUsage]
from . import _self_injecting  # pyright: ignore[reportPrivateUsage]
from . import _types  # pyright: ignore[reportPrivateUsage]
from . import _visitor
from . import abc as alluka

if typing.TYPE_CHECKING:
    from typing_extensions import Self


_T = typing.TypeVar("_T")


_AnyCoro = collections.Coroutine[typing.Any, typing.Any, typing.Any]
_CallbackSigT = typing.TypeVar("_CallbackSigT", bound=alluka.CallbackSig[typing.Any])
_DefaultT = typing.TypeVar("_DefaultT")
_SyncCallbackSigT = typing.TypeVar("_SyncCallbackSigT", bound=collections.Callable[..., typing.Any])

_TypeT = type[_T]


class _NoDefaultEnum(enum.Enum):
    VALUE = object()


_NO_VALUE: typing.Literal[_NoDefaultEnum.VALUE] = _NoDefaultEnum.VALUE
_NoValueOr = typing.Union[_T, typing.Literal[_NoDefaultEnum.VALUE]]


@typing.overload
def inject(*, callback: alluka.CallbackSig[_T]) -> _T: ...


@typing.overload
def inject(*, type: _TypeT[_T]) -> _T:  # noqa: A002
    ...


@typing.overload
def inject(*, type: typing.Any = None) -> typing.Any:  # noqa: A002
    ...


def inject(
    *, callback: typing.Optional[alluka.CallbackSig[_T]] = None, type: typing.Any = None  # noqa: A002
) -> typing.Any:
    """Declare a keyword-argument as requiring an injected dependency.

    This may be assigned to an argument's default value to declare injection
    or as a part of its Annotated metadata.

    !!! note
        If neither `type` nor `callback` is provided, an injected type
        will be inferred from the argument's annotation.

    Examples
    --------
    ```py
    async def callback(
        # Here we require an implementation of the type `Component` to be
        # injected.
        injected_type: Component = alluka.inject(type=Component)
        # Here we inject an out-of-scope callback which itself is taking
        # advantage of type injectioallukan.
        callback_result: ResultT = alluka.inject(callback=injected_callback)
    ) -> None:
        raise NotImplementedError

    ...
    # where client is an `alluka.Client` instance.
    result = await client.call_with_async_di(callback)
    ```

    Parameters
    ----------
    callback
        The callback to use to resolve the dependency.

        If this callback has no type dependencies then this will still work
        without an injection context but this can be overridden using
        `alluka.abc.Client.set_callback_override`.
    type
        The type of the dependency to resolve.

        If a union (e.g. `typing.Union[A, B]`, `A | B`, `typing.Optional[A]`)
        is passed for `type` then each type in the union will be tried
        separately rather than the literal type, allowing for resolving
        `A | B` to the value set by `set_type_dependency(B, ...)`.

        If a union has `None` as one of its types (including `Optional[T]`)
        then `None` will be passed for the parameter if none of the types could
        be resolved using the linked client.

    Raises
    ------
    ValueError
        If both `type` and `callback` are provided.
    """
    return typing.cast("_T", _types.InjectedDescriptor(callback=callback, type=type))


class Client(alluka.Client):
    """Standard implementation of a dependency injection client.

    This is used to track type dependencies and execute callbacks.
    """

    __slots__ = (
        "_callback_overrides",
        "_descriptors",
        "_introspect_annotations",
        "_make_context",
        "_type_dependencies",
    )

    def __init__(self, *, introspect_annotations: bool = True) -> None:
        """Initialise an injector client."""
        self._callback_overrides: dict[alluka.CallbackSig[typing.Any], alluka.CallbackSig[typing.Any]] = {}
        # TODO: this forces objects to have a __weakref__ attribute,
        # and also hashability (so hash and eq or neither), do we want to
        # keep with this behaviour or document it?
        self._descriptors: weakref.WeakKeyDictionary[
            alluka.CallbackSig[typing.Any], dict[str, _types.InjectedTuple]
        ] = weakref.WeakKeyDictionary()
        self._introspect_annotations = introspect_annotations
        static_context = _context.Context(self)
        self._make_context: collections.Callable[[Self], alluka.Context] = lambda _: static_context
        self._type_dependencies: dict[type[typing.Any], typing.Any] = {alluka.Client: self, Client: self}

    def _build_descriptors(self, callback: alluka.CallbackSig[typing.Any], /) -> dict[str, _types.InjectedTuple]:
        try:
            return self._descriptors[callback]

        except KeyError:
            pass

        # TODO: introspect_annotations=self._introspect_annotations
        descriptors = self._descriptors[callback] = _visitor.Callback(callback).accept(_visitor.ParameterVisitor())
        return descriptors

    def set_make_context(self, make_context: collections.Callable[[Self], alluka.Context], /) -> Self:
        """Set the callback used to make DI contexts for this client.

        Parameters
        ----------
        make_context
            Callback to use to create DI contexts.

        Returns
        -------
        Self
            The client to enable chained calls.
        """
        self._make_context = make_context
        return self

    def make_context(self) -> alluka.Context:
        # <<inherited docstring from alluka.abc.Client>>.
        return self._make_context(self)

    @typing_extensions.deprecated("Use .auto_inject")
    def as_async_self_injecting(
        self, callback: _CallbackSigT, /
    ) -> alluka.AsyncSelfInjecting[_CallbackSigT]:  # pyright: ignore[reportDeprecated]
        # <<inherited docstring from alluka.abc.Client>>.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            return _self_injecting.AsyncSelfInjecting(self, callback)  # pyright: ignore[reportDeprecated]

    @typing_extensions.deprecated("Use .auto_inject_async")
    def as_self_injecting(
        self, callback: _SyncCallbackSigT, /
    ) -> alluka.SelfInjecting[_SyncCallbackSigT]:  # pyright: ignore[reportDeprecated]
        # <<inherited docstring from alluka.abc.Client>>.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            return _self_injecting.SelfInjecting(self, callback)  # pyright: ignore[reportDeprecated]

    @typing.overload
    def call_with_ctx(
        self,
        ctx: alluka.Context,
        callback: collections.Callable[..., _AnyCoro],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.NoReturn: ...

    @typing.overload
    def call_with_ctx(
        self, ctx: alluka.Context, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T: ...

    def call_with_ctx(
        self, ctx: alluka.Context, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T:
        # <<inherited docstring from alluka.abc.Client>>.
        descriptors = self._build_descriptors(callback)
        if descriptors:
            # This prioritises passed **kwargs over the injected dependencies.
            kwargs = {n: v.resolve(ctx) for n, (_, v) in descriptors.items()} | kwargs

        result = callback(*args, **kwargs)
        if asyncio.iscoroutine(result):
            raise _errors.SyncOnlyError

        return result

    async def call_with_ctx_async(
        self, ctx: alluka.Context, callback: alluka.CallbackSig[_T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T:
        # <<inherited docstring from alluka.abc.Client>>.
        if descriptors := self._build_descriptors(callback):
            # Pyright currently doesn't support `is` for narrowing tuple types like this
            # This prioritises passed **kwargs over the injected dependencies.
            kwargs = {
                n: v[1].resolve(ctx) if v[0] == _types.InjectedTypes.TYPE else await v[1].resolve_async(ctx)
                for n, v in descriptors.items()
            } | kwargs

        result = callback(*args, **kwargs)
        if asyncio.iscoroutine(result):
            return typing.cast("_T", await result)

        return typing.cast("_T", result)  # noqa: ASYNC910  # This is ignored for performance's sake.

    def set_type_dependency(self, type_: type[_T], value: _T, /) -> Self:
        # <<inherited docstring from alluka.abc.Client>>.
        self._type_dependencies[type_] = value
        return self

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /) -> _T: ...

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    def get_type_dependency(
        self, type_: type[_T], /, *, default: _NoValueOr[_DefaultT] = _NO_VALUE
    ) -> typing.Union[_T, _DefaultT]:
        # <<inherited docstring from alluka.abc.Client>>.
        result = self._type_dependencies.get(type_, default)

        if result is _NO_VALUE:
            raise KeyError

        return result

    def remove_type_dependency(self, type_: type[typing.Any], /) -> Self:
        # <<inherited docstring from alluka.abc.Client>>.
        del self._type_dependencies[type_]
        return self

    def set_callback_override(self, callback: alluka.CallbackSig[_T], override: alluka.CallbackSig[_T], /) -> Self:
        # <<inherited docstring from alluka.abc.Client>>.
        self._callback_overrides[callback] = override
        return self

    def get_callback_override(self, callback: alluka.CallbackSig[_T], /) -> typing.Optional[alluka.CallbackSig[_T]]:
        # <<inherited docstring from alluka.abc.Client>>.
        return self._callback_overrides.get(callback)

    def remove_callback_override(self, callback: alluka.CallbackSig[_T], /) -> Self:
        # <<inherited docstring from alluka.abc.Client>>.
        del self._callback_overrides[callback]
        return self



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# pyright: reportIncompatibleMethodOverride=none

import typing
import warnings

import mock
import pytest

import alluka


def test_inject():
    descriptor = alluka.inject()

    assert descriptor.type is None
    assert descriptor.callback is None


def test_inject_when_type():
    mock_type = mock.Mock()

    descriptor = alluka.inject(type=mock_type)

    assert descriptor.type is mock_type
    assert descriptor.callback is None


def test_inject_when_callback():
    mock_callback = mock.Mock()

    descriptor = alluka.inject(callback=mock_callback)

    assert descriptor.type is None
    assert descriptor.callback is mock_callback


def test_inject_when_both_callback_and_type():
    with pytest.raises(ValueError, match="Only one of `callback` or `type` can be specified"):
        alluka.inject(type=mock.Mock(), callback=mock.Mock())  # type: ignore


class TestClient:
    def test_make_context(self):
        client = alluka.Client()

        context = client.make_context()

        assert isinstance(context, alluka.Context)
        assert context.injection_client is client

    def test_make_context_when_overridden(self):
        mock_maker = mock.Mock()
        client = alluka.Client().set_make_context(mock_maker)

        assert client.make_context() is mock_maker.return_value
        mock_maker.assert_called_once_with(client)

    def test_as_async_self_injecting(self):
        mock_callback = mock.Mock()
        client = alluka.Client()

        with pytest.warns(DeprecationWarning):
            result = client.as_async_self_injecting(mock_callback)  # pyright: ignore[reportDeprecated]

        assert isinstance(result, alluka.AsyncSelfInjecting)  # pyright: ignore[reportDeprecated]
        assert result._callback is mock_callback
        assert result._client is client

    def test_as_self_injecting(self):
        mock_callback = mock.Mock()
        client = alluka.Client()

        with pytest.warns(DeprecationWarning):
            result = client.as_self_injecting(mock_callback)  # pyright: ignore[reportDeprecated]

        assert isinstance(result, alluka.SelfInjecting)  # pyright: ignore[reportDeprecated]
        assert result._callback is mock_callback
        assert result._client is client

    def test_call_with_di(self):
        class MockType1: ...

        class MockType2: ...

        class MockType3: ...

        mock_value_1 = MockType1()
        mock_value_2 = MockType2()
        mock_value_3 = MockType3()
        mock_callback = mock.Mock()
        mock_override = mock.Mock()

        def callback(
            value_1: int,
            value_2: str,
            value_3: alluka.Injected[MockType1],
            value_4: alluka.Injected[typing.Union[int, str, None]],
            other_result: typing.Annotated[int, alluka.inject(callback=lambda: "no you")],
            overridden: typing.Annotated[int, alluka.inject(callback=mock_callback)],
            bart_man: alluka.Injected[int] = 123,
            meow: MockType3 = alluka.inject(type=MockType3),
            nyaa: typing.Optional[bool] = alluka.inject(type=typing.Optional[bool]),
            result: typing.Optional[str] = alluka.inject(callback=lambda: "hi"),
        ) -> str:
            assert value_1 == 43234
            assert value_2 == "nyaa"
            assert value_3 is mock_value_1
            assert value_4 is None
            assert other_result == "no you"
            assert overridden is mock_override.return_value
            assert bart_man == 123
            assert meow is mock_value_3
            assert nyaa is None
            assert result == "hi"
            return "ok"

        client = (
            alluka.Client()
            .set_type_dependency(MockType1, mock_value_1)
            .set_type_dependency(MockType2, mock_value_2)
            .set_type_dependency(MockType3, mock_value_3)
            .set_callback_override(mock_callback, mock_override)
        )

        result = client.call_with_di(callback, 43234, value_2="nyaa")

        assert result == "ok"

    def test_call_with_di_when_type_not_found(self):
        class MockType: ...

        def callback(value: alluka.Injected[MockType]) -> typing.NoReturn:
            raise NotImplementedError

        client = alluka.Client()

        with pytest.raises(alluka.MissingDependencyError):
            client.call_with_di(callback)

    def test_call_with_di_when_type_not_found_when_async_callback(self):
        async def callback(value: alluka.Injected[int]) -> typing.NoReturn:
            raise NotImplementedError

        client = alluka.Client().set_type_dependency(int, 123)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)

            with pytest.raises(alluka.SyncOnlyError):
                client.call_with_di(callback)

    def test_call_with_di_when_type_not_found_when_async_dependency(self):
        class MockType: ...

        def callback(
            value: alluka.Injected[MockType], dep: int = alluka.inject(callback=mock.AsyncMock())
        ) -> typing.NoReturn:
            raise NotImplementedError

        client = alluka.Client().set_type_dependency(MockType, MockType())

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)

            with pytest.raises(alluka.SyncOnlyError):
                client.call_with_di(callback)

    @pytest.mark.anyio
    async def test_call_with_async_di(self):
        class MockType1: ...

        class MockType2: ...

        class MockType3: ...

        mock_value_1 = MockType1()
        mock_value_2 = MockType2()
        mock_value_3 = MockType3()
        mock_callback = mock.Mock()
        mock_override = mock.AsyncMock()
        mock_other_callback = mock.AsyncMock()
        mock_other_override = mock.Mock()

        async def callback(
            foo: int,
            bar: str,
            bam: alluka.Injected[MockType1],
            baz: alluka.Injected[typing.Union[int, str, None]],
            other_result: typing.Annotated[int, alluka.inject(callback=lambda: "no you")],
            overridden: typing.Annotated[int, alluka.inject(callback=mock_callback)],
            other_overridden: typing.Any = alluka.inject(callback=mock_other_callback),
            bart_man: alluka.Injected[int] = 123,
            meow: MockType3 = alluka.inject(type=MockType3),
            nyaa: typing.Optional[bool] = alluka.inject(type=typing.Optional[bool]),
            result: typing.Optional[str] = alluka.inject(callback=lambda: "hi"),
        ) -> str:
            assert foo == 43234
            assert bar == "nyaa"
            assert bam is mock_value_1
            assert baz is None
            assert other_result == "no you"
            assert overridden is mock_override.return_value
            assert other_overridden is mock_other_override.return_value
            assert bart_man == 123
            assert meow is mock_value_3
            assert nyaa is None
            assert result == "hi"
            return "ok"

        client = (
            alluka.Client()
            .set_type_dependency(MockType1, mock_value_1)
            .set_type_dependency(MockType2, mock_value_2)
            .set_type_dependency(MockType3, mock_value_3)
            .set_callback_override(mock_callback, mock_override)
            .set_callback_override(mock_other_callback, mock_other_override)
        )

        result = await client.call_with_async_di(callback, 43234, bar="nyaa")

        assert result == "ok"

    @pytest.mark.anyio
    async def test_call_with_async_di_when_type_not_found(self):
        class MockType: ...

        def callback(value: alluka.Injected[MockType]) -> typing.NoReturn:
            raise NotImplementedError

        client = alluka.Client()

        with pytest.raises(alluka.MissingDependencyError):
            await client.call_with_async_di(callback)

    def test_set_type_dependency_when_not_found(self):
        mock_type: typing.Any = mock.Mock()
        client = alluka.Client()

        with pytest.raises(KeyError):
            client.get_type_dependency(mock_type)

    def test_get_type_dependency_when_not_found_and_default(self):
        mock_type: typing.Any = mock.Mock()
        default = object()
        client = alluka.Client()

        result = client.get_type_dependency(mock_type, default=default)

        assert result is default

    def test_remove_type_dependency(self):
        mock_type: typing.Any = mock.Mock()
        client = alluka.Client()
        client.set_type_dependency(mock_type, mock.Mock())

        result = client.remove_type_dependency(mock_type)

        assert result is client

        with pytest.raises(KeyError):
            assert client.get_type_dependency(mock_type)

    def test_remove_type_dependency_when_not_set(self):
        mock_type: typing.Any = mock.Mock()
        client = alluka.Client()

        with pytest.raises(KeyError):
            client.remove_type_dependency(mock_type)

    def test_set_callback_override(self):
        mock_callback = mock.Mock()
        mock_override = mock.Mock()
        client = alluka.Client()

        result = client.set_callback_override(mock_callback, mock_override)

        assert result is client
        assert client.get_callback_override(mock_callback) is mock_override

    def test_get_callback_override(self):
        client = alluka.Client()

        assert client.get_callback_override(mock.Mock()) is None

    def test_remove_callback_override(self):
        mock_callback = mock.Mock()
        client = alluka.Client()
        client.set_callback_override(mock_callback, mock.Mock())

        result = client.remove_callback_override(mock_callback)

        assert result is client
        assert client.get_callback_override(mock_callback) is None

    def test_remove_callback_override_when_not_set(self):
        mock_callback = mock.Mock()
        client = alluka.Client()

        with pytest.raises(KeyError):
            client.remove_callback_override(mock_callback)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""The standard visitor and nodes used to parse function parameters."""
from __future__ import annotations

__all__: list[str] = ["Callback", "ParameterVisitor"]

import abc
import sys
import types
import typing

from . import _types  # pyright: ignore[reportPrivateUsage]
from ._vendor import inspect

if typing.TYPE_CHECKING:
    from collections import abc as collections


if sys.version_info >= (3, 10):
    _UnionTypes = frozenset((typing.Union, types.UnionType))
    _NoneType = types.NoneType

else:
    _UnionTypes = frozenset((typing.Union,))
    _NoneType = type(None)


class Node(abc.ABC):
    """Represents a node which may hold parameter injection information."""

    __slots__ = ()

    @abc.abstractmethod
    def accept(self, visitor: ParameterVisitor, /) -> typing.Optional[_types.InjectedTuple]: ...


class Annotation(Node):
    """Node which represent's a parameter's type-hint."""

    __slots__ = ("_callback", "_name", "_raw_annotation")

    def __init__(self, callback: Callback, name: str, /):
        self._callback = callback
        self._name = name
        self._raw_annotation = callback.parameters[name].annotation

    @property
    def callback(self) -> Callback:
        return self._callback

    @property
    def name(self) -> str:
        return self._name

    def accept(self, visitor: ParameterVisitor, /) -> typing.Optional[_types.InjectedTuple]:
        return visitor.visit_annotation(self)


class Callback:
    """Represents a callback being scanned for DI."""

    __slots__ = ("_callback", "_resolved", "_signature")

    def __init__(self, callback: collections.Callable[..., typing.Any], /) -> None:
        self._callback: collections.Callable[..., typing.Any] = callback
        self._resolved = False
        try:
            self._signature: typing.Optional[inspect.Signature] = inspect.signature(callback)
        except ValueError:  # If we can't inspect it then we have to assume this is a NO
            # As a note, this fails on some "signature-less" builtin functions/types like str.
            self._signature = None

    @property
    def parameters(self) -> collections.Mapping[str, inspect.Parameter]:
        return self._signature.parameters if self._signature else {}

    def accept(self, visitor: ParameterVisitor, /) -> dict[str, _types.InjectedTuple]:
        return visitor.visit_callback(self)

    def resolve_annotation(self, name: str, /) -> _types.UndefinedOr[typing.Any]:
        if self._signature is None:
            return _types.UNDEFINED

        parameter = self._signature.parameters[name]
        if parameter.annotation is inspect.Parameter.empty:
            return _types.UNDEFINED

        # TODO: do we want to return UNDEFINED if it was resolved to a string?
        if not self._resolved and isinstance(parameter.annotation, str):
            self._signature = inspect.signature(self._callback, eval_str=True)
            self._resolved = True
            return self.resolve_annotation(name)

        return parameter.annotation


class Default(Node):
    """Node which represent a parameter's default value."""

    __slots__ = ("_callback", "_default", "_name")

    def __init__(self, callback: Callback, name: str, /) -> None:
        self._callback = callback
        self._default = callback.parameters[name].default
        self._name = name

    @property
    def callback(self) -> Callback:
        return self._callback

    @property
    def default(self) -> typing.Any:
        return self._default

    @property
    def is_empty(self) -> bool:
        return self._default is inspect.Parameter.empty

    @property
    def name(self) -> str:
        return self._name

    def accept(self, visitor: ParameterVisitor, /) -> typing.Optional[_types.InjectedTuple]:
        return visitor.visit_default(self)


class ParameterVisitor:
    """Visitor class for parsing a callback for injected parameters."""

    __slots__ = ()

    _NODES: list[collections.Callable[[Callback, str], Node]] = [Default, Annotation]

    def _parse_type(
        self, type_: typing.Any, *, default: _types.UndefinedOr[typing.Any] = _types.UNDEFINED
    ) -> _types.InjectedTuple:
        if typing.get_origin(type_) not in _UnionTypes:
            return (_types.InjectedTypes.TYPE, _types.InjectedType(type_, [type_], default=default))

        sub_types = list(typing.get_args(type_))
        try:
            sub_types.remove(_NoneType)
        except ValueError:
            return (_types.InjectedTypes.TYPE, _types.InjectedType(type_, sub_types, default=default))

        # Explicitly defined defaults take priority over implicit defaults.
        default = None if default is _types.UNDEFINED else default
        return (_types.InjectedTypes.TYPE, _types.InjectedType(type_, sub_types, default=default))

    def _annotation_to_type(
        self, value: typing.Any, /, default: _types.UndefinedOr[typing.Any] = _types.UNDEFINED
    ) -> _types.InjectedTuple:
        if typing.get_origin(value) is typing.Annotated:
            args = typing.get_args(value)
            # The first "type" arg of annotated will always be flatterned to a type.
            # so we don't have to deal with Annotated nesting".
            value = args[0]

        return self._parse_type(value, default=default)

    def visit_annotation(self, annotation: Annotation, /) -> typing.Optional[_types.InjectedTuple]:
        value = annotation.callback.resolve_annotation(annotation.name)
        default = annotation.callback.parameters[annotation.name].default
        if default is inspect.Parameter.empty:
            default = _types.UNDEFINED

        if typing.get_origin(value) is not typing.Annotated:
            return None

        args = typing.get_args(value)
        if _types.InjectedTypes.TYPE in args:
            return self._annotation_to_type(args[0], default=default)

        arg: typing.Union[_types.InjectedDescriptor[typing.Any], typing.Any]
        for arg in args:
            if not isinstance(arg, _types.InjectedDescriptor):
                continue

            if arg.callback:
                return (_types.InjectedTypes.CALLBACK, _types.InjectedCallback(arg.callback))

            if arg.type:
                return self._parse_type(arg.type, default=default)

            return self._annotation_to_type(args[0], default=default)

        return None  # MyPy

    def visit_callback(self, callback: Callback, /) -> dict[str, _types.InjectedTuple]:
        results: dict[str, _types.InjectedTuple] = {}
        for name, value in callback.parameters.items():
            for node in self._NODES:
                result = node(callback, name).accept(self)
                if not result:
                    continue

                if value.kind is value.POSITIONAL_ONLY:
                    raise ValueError("Injected positional only arguments are not supported")

                results[name] = result
                break

        return results

    def visit_default(self, value: Default, /) -> typing.Optional[_types.InjectedTuple]:
        if value.is_empty or not isinstance(value.default, _types.InjectedDescriptor):
            return None  # MyPy

        descriptor: _types.InjectedDescriptor[typing.Any] = value.default
        if descriptor.callback is not None:
            return (_types.InjectedTypes.CALLBACK, _types.InjectedCallback(descriptor.callback))

        if descriptor.type is not None:
            return self._parse_type(descriptor.type)

        if (annotation := value.callback.resolve_annotation(value.name)) is _types.UNDEFINED:
            raise ValueError(f"Could not resolve type for parameter {value.name!r} with no annotation")

        return self._annotation_to_type(annotation)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Alluka's abstract interfaces."""

# pyright: reportOverlappingOverload=warning

from __future__ import annotations

__all__: list[str] = ["AsyncSelfInjecting", "CallbackSig", "Client", "Context", "SelfInjecting"]

import abc
import enum
import functools
import typing
from collections import abc as collections

import typing_extensions

if typing.TYPE_CHECKING:
    from typing_extensions import Self

    _P = typing_extensions.ParamSpec("_P")

_T = typing.TypeVar("_T")
_CoroT = collections.Coroutine[typing.Any, typing.Any, _T]
_CallbackT = typing.TypeVar("_CallbackT", bound="CallbackSig[typing.Any]")
_DefaultT = typing.TypeVar("_DefaultT")
_SyncCallbackT = typing.TypeVar("_SyncCallbackT", bound=collections.Callable[..., typing.Any])


class _NoDefaultEnum(enum.Enum):
    VALUE = object()


_NO_VALUE: typing.Literal[_NoDefaultEnum.VALUE] = _NoDefaultEnum.VALUE
_NoValueOr = typing.Union[_T, typing.Literal[_NoDefaultEnum.VALUE]]


CallbackSig = collections.Callable[..., typing.Union[_CoroT[_T], _T]]
"""Type-hint of a injector callback.

!!! note
    Dependency dependency injection is recursively supported, meaning that the
    keyword arguments for a dependency callback may also ask for dependencies
    themselves.

This may either be a sync or asyc function with dependency injection being
available for the callback's keyword arguments but dynamically returning either
a coroutine or raw value may lead to errors.

Dependent on the context positional arguments may also be proivded.
"""


class Client(abc.ABC):
    """Abstract interface of a dependency injection client."""

    __slots__ = ()

    @abc.abstractmethod
    def make_context(self) -> Context:
        """Create a dependency injection context.

        Returns
        -------
        alluka.abc.Context
            The created DI context, bound to the current client.
        """

    @abc.abstractmethod
    @typing_extensions.deprecated("Use .auto_inject_async")
    def as_async_self_injecting(
        self, callback: _CallbackT, /
    ) -> AsyncSelfInjecting[_CallbackT]:  # pyright: ignore[reportDeprecated]
        """Deprecated callback for making async functions auto-inject.

        Use [Client.auto_inject_async][alluka.abc.Client.auto_inject_async].
        """

    @abc.abstractmethod
    @typing_extensions.deprecated("Use .auto_inject")
    def as_self_injecting(
        self, callback: _SyncCallbackT, /
    ) -> SelfInjecting[_SyncCallbackT]:  # pyright: ignore[reportDeprecated]
        """Deprecated callback for making functions auto-inject.

        Use [Client.auto_inject][alluka.abc.Client.auto_inject].
        """

    def auto_inject(self, callback: collections.Callable[_P, _T], /) -> collections.Callable[_P, _T]:
        """Wrap a function to make calls to it always inject dependencies.

        Examples
        --------
        ```py
        @client.auto_inject
        def callback(dep: Injected[Type]) -> None:
            ...

        callback()  # The requested dependencies will be passed.
        ```

        Parameters
        ----------
        callback
            The callback to wrap with DI.

        Returns
        -------
        collections.Callable
            The wrapped auto injecting callback.
        """

        @functools.wraps(callback)
        def wrapped_callback(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            return self.call_with_di(callback, *args, **kwargs)

        return wrapped_callback

    def auto_inject_async(
        self, callback: collections.Callable[_P, _CoroT[_T]], /
    ) -> collections.Callable[_P, _CoroT[_T]]:
        """Wrap an async function to make calls to it always inject dependencies.

        Examples
        --------
        ```py
        @client.auto_inject_async
        async def callback(dep: Injected[Type]) -> None:
            ...

        await callback()  # The requested dependencies will be passed.
        ```

        Parameters
        ----------
        callback
            The callback to wrap with DI.

        Returns
        -------
        collections.Callable
            The wrapped auto injecting callback.
        """

        @functools.wraps(callback)
        def wrapped_callback(*args: _P.args, **kwargs: _P.kwargs) -> _CoroT[_T]:
            return self.call_with_async_di(callback, *args, **kwargs)

        return wrapped_callback

    @typing.overload
    def call_with_di(
        self, callback: collections.Callable[..., _CoroT[typing.Any]], *args: typing.Any, **kwargs: typing.Any
    ) -> typing.NoReturn: ...

    @typing.overload
    def call_with_di(self, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any) -> _T: ...

    def call_with_di(self, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any) -> _T:
        """Call a function with sync dependency injection.

        Parameters
        ----------
        callback
            The callback to call.

            This must be sync.
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback.

        Returns
        -------
        _T
            The result of the callback.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        alluka.SyncOnlyError
            If the callback or any of its callback dependencies are async.
        """
        return self.make_context().call_with_di(callback, *args, **kwargs)

    @typing.overload
    @abc.abstractmethod
    def call_with_ctx(
        self,
        ctx: Context,
        callback: collections.Callable[..., _CoroT[typing.Any]],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.NoReturn: ...

    @typing.overload
    @abc.abstractmethod
    def call_with_ctx(
        self, ctx: Context, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T: ...

    @abc.abstractmethod
    def call_with_ctx(
        self, ctx: Context, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T:
        """Call a function with an existing DI context.

        Parameters
        ----------
        ctx
            The DI context to call the callback with.

            This will be used for scoped type injection.
        callback
            The callback to call.

            This must be sync.
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback.

        Returns
        -------
        _T
            The result of the callback.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        alluka.SyncOnlyError
            If the callback or any of its callback dependencies are async.
        """

    async def call_with_async_di(self, callback: CallbackSig[_T], *args: typing.Any, **kwargs: typing.Any) -> _T:
        """Call a function with async dependency injection.

        Parameters
        ----------
        callback
            The callback to call.

            This may be sync or async.
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback.

        Returns
        -------
        _T
            The result of the callback.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        alluka.SyncOnlyError
            If the callback or any of its callback dependencies are async.
        """
        return await self.make_context().call_with_async_di(callback, *args, **kwargs)

    @abc.abstractmethod
    async def call_with_ctx_async(
        self, ctx: Context, callback: CallbackSig[_T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T:
        """Asynchronously call a function with a pre-existing DI context.

        Parameters
        ----------
        ctx
            The DI context to call the callback with.

            This will be used for scoped type injection.
        callback
            The callback to call.

            This may be sync or async.
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback.

        Returns
        -------
        _T
            The result of the callback.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        """

    @abc.abstractmethod
    def set_type_dependency(self, type_: type[_T], value: _T, /) -> Self:
        """Set the value for a type dependency.

        Parameters
        ----------
        type_
            The type of the dependency to add an implementation for.
        value
            The value of the dependency.

        Returns
        -------
        Self
            The client instance to allow chaining.
        """

    @typing.overload
    @abc.abstractmethod
    def get_type_dependency(self, type_: type[_T], /) -> _T: ...

    @typing.overload
    @abc.abstractmethod
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    @abc.abstractmethod
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT = ...) -> typing.Union[_T, _DefaultT]:
        """Get the implementation for an injected type.

        Parameters
        ----------
        type_
            The associated type.
        default
            The default value to return if the type is not implemented.

        Returns
        -------
        _T | _DefaultT
            The resolved type if found.

            If the type isn't implemented then the value of `default`
            will be returned if it is provided.

        Raises
        ------
        KeyError
            If no dependency was found when no default was provided.
        """

    @abc.abstractmethod
    def remove_type_dependency(self, type_: type[typing.Any], /) -> Self:
        """Remove a type dependency.

        Parameters
        ----------
        type_
            The associated type.

        Returns
        -------
        Self
            The client instance to allow chaining.

        Raises
        ------
        KeyError
            If `type` is not registered.
        """

    @abc.abstractmethod
    def set_callback_override(self, callback: CallbackSig[_T], override: CallbackSig[_T], /) -> Self:
        """Override a specific injected callback.

        Parameters
        ----------
        callback
            The injected callback to override.
        override
            The callback to use instead.

        Returns
        -------
        Self
            The client instance to allow chaining.
        """

    @abc.abstractmethod
    def get_callback_override(self, callback: CallbackSig[_T], /) -> typing.Optional[CallbackSig[_T]]:
        """Get the override for a specific injected callback.

        Parameters
        ----------
        callback
            The injected callback to get the override for.

        Returns
        -------
        CallbackSig[_T] | None
            The override if found, else [None][].
        """

    @abc.abstractmethod
    def remove_callback_override(self, callback: CallbackSig[_T], /) -> Self:
        """Remove a callback override.

        Parameters
        ----------
        callback
            The injected callback to remove the override for.

        Returns
        -------
        Self
            The client instance to allow chaining.

        Raises
        ------
        KeyError
            If no override is found for the callback.
        """


class Context(abc.ABC):
    """Abstract interface of an injection context."""

    __slots__ = ()

    @property
    @abc.abstractmethod
    def injection_client(self) -> Client:
        """Injection client this context is bound to."""

    def cache_result(self, callback: CallbackSig[_T], value: _T, /) -> None:
        """Cache the result of a callback within the scope of this context.

        Whether this does anything or is a noop is implementation detail.

        Parameters
        ----------
        callback
            The callback to cache the result of.
        value
            The value to cache.
        """

    @typing.overload
    def call_with_di(
        self, callback: collections.Callable[..., _CoroT[typing.Any]], *args: typing.Any, **kwargs: typing.Any
    ) -> typing.NoReturn: ...

    @typing.overload
    def call_with_di(self, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any) -> _T: ...

    def call_with_di(self, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any) -> _T:
        """Call a function with the current DI context.

        Parameters
        ----------
        callback
            The callback to call.

            This must be sync.
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback.

        Returns
        -------
        _T
            The result of the callback.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        alluka.SyncOnlyError
            If the callback or any of its callback dependencies are async.
        """
        return self.injection_client.call_with_ctx(self, callback, *args, **kwargs)

    async def call_with_async_di(self, callback: CallbackSig[_T], *args: typing.Any, **kwargs: typing.Any) -> _T:
        """Asynchronously call a function with the current DI context.

        Parameters
        ----------
        callback
            The callback to call.

            This may be sync or async.
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback.

        Returns
        -------
        _T
            The result of the callback.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        """
        return await self.injection_client.call_with_ctx_async(self, callback, *args, **kwargs)

    @typing.overload
    def get_cached_result(self, callback: CallbackSig[_T], /) -> _T: ...

    @typing.overload
    def get_cached_result(self, callback: CallbackSig[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    def get_cached_result(
        self, callback: CallbackSig[_T], /, *, default: _NoValueOr[_DefaultT] = _NO_VALUE
    ) -> typing.Union[_T, _DefaultT]:
        """Get the cached result of a callback.

        This will always raise/default for context implementations with no caching.

        Parameters
        ----------
        callback
            The callback to get the cached result of.
        default
            The default value to return if the callback is not cached.

        Returns
        -------
        _T | _DefaultT
            The cached result of the callback if found.

            If the callback's result hasn't been cached or caching isn't
            implementing then this will return the value of `default` if it
            is provided.

        Raises
        ------
        KeyError
            If no value was found when no default was provided.
        """
        if default is _NO_VALUE:
            raise KeyError

        return default

    @typing.overload
    @abc.abstractmethod
    def get_type_dependency(self, type_: type[_T], /) -> _T: ...

    @typing.overload
    @abc.abstractmethod
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    @abc.abstractmethod
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT = ...) -> typing.Union[_T, _DefaultT]:
        """Get the implementation for an injected type.

        Unlike [Client.get_type_dependency][alluka.abc.Client.get_type_dependency],
        this method may also return context specific implementations of a type.

        Parameters
        ----------
        type_
            The associated type.
        default
            The default value to return if the type is not implemented.

        Returns
        -------
        _T | _DefaultT
            The resolved type if found.

            If the type isn't implemented then the value of `default`
            will be returned if it is provided.

        Raises
        ------
        KeyError
            If no dependency was found when no default was provided.
        """


@typing_extensions.deprecated("Use Client.auto_inject_async")
class AsyncSelfInjecting(abc.ABC, typing.Generic[_CallbackT]):
    """Deprecated interface of a class for marking async functions as self-injecting.

    Use [Client.auto_inject_async][alluka.abc.Client.auto_inject_async].
    """

    __slots__ = ()

    @property
    @abc.abstractmethod
    def callback(self) -> _CallbackT:
        """The callback this wraps."""

    @typing.overload
    @abc.abstractmethod
    async def __call__(
        self: AsyncSelfInjecting[collections.Callable[..., _CoroT[_T]]],  # pyright: ignore[reportDeprecated]
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> _T: ...

    @typing.overload
    @abc.abstractmethod
    async def __call__(
        self: AsyncSelfInjecting[collections.Callable[..., _T]],  # pyright: ignore[reportDeprecated]
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> _T: ...

    @abc.abstractmethod
    async def __call__(
        self: typing.Union[
            AsyncSelfInjecting[collections.Callable[..., _T]],  # pyright: ignore[reportDeprecated]
            AsyncSelfInjecting[collections.Callable[..., _CoroT[_T]]],  # pyright: ignore[reportDeprecated]
        ],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> _T:
        """Call this with the provided arguments and any injected arguments.

        Parameters
        ----------
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback alongside injected arguments.

        Returns
        -------
        _T
            The callback's result.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        """


@typing_extensions.deprecated("Use Client.auto_inject")
class SelfInjecting(abc.ABC, typing.Generic[_SyncCallbackT]):
    """Deprecated interface of a class for marking functions as self-injecting.

    Use [Client.auto_inject][alluka.abc.Client.auto_inject].
    """

    __slots__ = ()

    @property
    @abc.abstractmethod
    def callback(self) -> _SyncCallbackT:
        """The callback this wraps."""

    @abc.abstractmethod
    def __call__(
        self: SelfInjecting[collections.Callable[..., _T]],  # pyright: ignore[reportDeprecated]
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> _T:
        """Call this callback with the provided arguments + injected arguments.

        Parameters
        ----------
        *args
            Positional arguments to pass to the callback.
        **kwargs
            Keyword arguments to pass to the callback.

        Returns
        -------
        _T
            The callback's result.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's required type dependencies aren't implemented
            by the client.
        alluka.SyncOnlyError
            If the callback or any of its callback dependencies are async.
        """



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportDeprecated=none

from __future__ import annotations

__all__: list[str] = ["AsyncSelfInjecting", "SelfInjecting"]

import typing
import warnings
from collections import abc as collections

import typing_extensions

from . import abc as alluka

_CallbackSigT = typing.TypeVar("_CallbackSigT", bound=alluka.CallbackSig[typing.Any])
_SyncCallbackT = typing.TypeVar("_SyncCallbackT", bound=collections.Callable[..., typing.Any])
_T = typing.TypeVar("_T")
_CoroT = collections.Coroutine[typing.Any, typing.Any, _T]


with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)

    @typing_extensions.deprecated("Use Client.auto_inject_async")
    class AsyncSelfInjecting(alluka.AsyncSelfInjecting[_CallbackSigT]):
        """Deprecated class for marking async functions as self-injecting.

        Use [Client.auto_inject_async][alluka.abc.Client.auto_inject_async].
        """

        __slots__ = ("_callback", "_client")

        def __init__(self, client: alluka.Client, callback: _CallbackSigT, /) -> None:
            """Initialise a self injecting callback.

            Parameters
            ----------
            client
                The injection client to use to resolve dependencies.
            callback : alluka.abc.CallbackSig
                The callback to make self-injecting.

                This may be sync or async.

            Raises
            ------
            ValueError
                If `callback` has any injected arguments which can only be passed
                positionally.
            """
            self._callback = callback
            self._client = client

        @typing.overload
        async def __call__(
            self: AsyncSelfInjecting[collections.Callable[..., _CoroT[_T]]], *args: typing.Any, **kwargs: typing.Any
        ) -> _T: ...

        @typing.overload
        async def __call__(
            self: AsyncSelfInjecting[collections.Callable[..., _T]], *args: typing.Any, **kwargs: typing.Any
        ) -> _T: ...

        async def __call__(
            self: typing.Union[
                AsyncSelfInjecting[collections.Callable[..., _T]],
                AsyncSelfInjecting[collections.Callable[..., _CoroT[_T]]],
            ],
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> _T:
            # <<inherited docstring from alluka.abc.AsyncSelfInjecting>>.
            return await self._client.call_with_async_di(self._callback, *args, **kwargs)

        @property
        def callback(self) -> _CallbackSigT:
            # <<inherited docstring from alluka.abc.AsyncSelfInjecting>>.
            return self._callback

    @typing_extensions.deprecated("Use Client.inject_async")
    class SelfInjecting(alluka.SelfInjecting[_SyncCallbackT]):
        """Deprecated class for marking functions as self-injecting.

        Use [Client.auto_inject][alluka.abc.Client.auto_inject].
        """

        __slots__ = ("_callback", "_client")

        def __init__(self, client: alluka.Client, callback: _SyncCallbackT, /) -> None:
            """Initialise a sync self injecting callback.

            Parameters
            ----------
            client
                The injection client to use to resolve dependencies.
            callback : collections.abc.Callable
                The callback to make self-injecting.

            Raises
            ------
            ValueError
                If `callback` has any injected arguments which can only be passed
                positionally.
            """
            self._callback = callback
            self._client = client

        def __call__(self: SelfInjecting[collections.Callable[..., _T]], *args: typing.Any, **kwargs: typing.Any) -> _T:
            # <<inherited docstring from alluka.abc.SelfInjecting>>.
            return self._client.call_with_di(self._callback, *args, **kwargs)

        @property
        def callback(self) -> _SyncCallbackT:
            # <<inherited docstring from alluka.abc.SelfInjecting>>.
            return self._callback



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# pyright: reportIncompatibleMethodOverride=none

import typing

import mock
import pytest

import alluka


class TestContext:
    def test_injection_client_property(self):
        mock_client = alluka.Client()
        ctx = alluka.Context(mock_client)

        assert ctx.injection_client is mock_client

    def test_get_type_dependency(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        mock_client = alluka.Client().set_type_dependency(mock_type, mock_value)
        ctx = alluka.Context(mock_client)

        result = ctx.get_type_dependency(mock_type)

        assert result is mock_value

    def test_get_type_dependency_when_type_is_context_abc(self):
        mock_client = alluka.Client().set_type_dependency(alluka.abc.Context, mock.Mock())
        ctx = alluka.Context(mock_client)

        result = ctx.get_type_dependency(alluka.abc.Context)

        assert result is ctx

    def test_get_type_dependency_with_default(self):
        default = object()
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        mock_client = alluka.Client().set_type_dependency(mock_type, mock_value)
        ctx = alluka.Context(mock_client)

        result = ctx.get_type_dependency(mock_type, default=default)

        assert result is mock_value

    def test_get_type_dependency_when_not_found_with_default(self):
        default = object()
        mock_type: typing.Any = mock.Mock()
        ctx = alluka.Context(alluka.Client())

        result = ctx.get_type_dependency(mock_type, default=default)

        assert result is default

    def test_get_type_dependency_when_not_found(self):
        mock_type: typing.Any = mock.Mock()
        ctx = alluka.Context(alluka.Client())

        with pytest.raises(KeyError):
            ctx.get_type_dependency(mock_type)


class TestCachingContext:
    def test_cache_result(self):
        mock_callback = mock.Mock()
        mock_result = mock.Mock()
        ctx = alluka.CachingContext(alluka.Client())
        ctx.cache_result(mock_callback, mock_result)

        assert ctx.get_cached_result(mock_callback) is mock_result

    def test_get_cached_result_when_not_found(self):
        ctx = alluka.CachingContext(alluka.Client())

        with pytest.raises(KeyError):
            ctx.get_cached_result(mock.Mock())

    def test_get_cached_result_when_not_found_with_default(self):
        ctx = alluka.CachingContext(alluka.Client())
        default = object()

        assert ctx.get_cached_result(mock.Mock(), default=default) is default


class TestOverridingContext:
    def test_from_client(self):
        client = mock.Mock(alluka.Client)
        expected_context = client.make_context.return_value

        ctx = alluka.OverridingContext.from_client(client)

        assert ctx.injection_client is expected_context.injection_client
        assert ctx.get_type_dependency(int, default=123) is expected_context.get_type_dependency.return_value
        expected_context.get_type_dependency.assert_called_once_with(int, default=123)
        client.make_context.assert_called_once_with()

    def test_injection_client_property(self):
        mock_context = mock.Mock()
        ctx = alluka.OverridingContext(mock_context)

        assert ctx._context is mock_context

    def test_get_cached_result(self):
        mock_context = mock.Mock()
        mock_callback = mock.Mock()
        ctx = alluka.OverridingContext(mock_context)

        result = ctx.get_cached_result(mock_callback, default=123321)

        assert result is mock_context.get_cached_result.return_value
        mock_context.get_cached_result.assert_called_once_with(mock_callback, default=123321)

    def test_get_cached_result_when_not_cached(self):
        mock_callback = mock.Mock()
        ctx = alluka.OverridingContext(alluka.Client().make_context())

        with pytest.raises(KeyError):
            ctx.get_cached_result(mock_callback)

    def test_get_cached_result_when_not_cached_with_default_set(self):
        mock_callback = mock.Mock()
        ctx = alluka.OverridingContext(alluka.Client().make_context())

        result = ctx.get_cached_result(mock_callback, default=4333)

        assert result == 4333

    def test_cache_result(self):
        mock_context = mock.Mock()
        mock_callback = mock.Mock()
        mock_value = mock.Mock()
        ctx = alluka.OverridingContext(mock_context)

        ctx.cache_result(mock_callback, mock_value)

        mock_context.cache_result.assert_called_once_with(mock_callback, mock_value)

    def test_get_type_dependency(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        context = alluka.Client().set_type_dependency(mock_type, mock_value).make_context()
        context = alluka.OverridingContext(context)

        assert context.get_type_dependency(mock_type) is mock_value

    def test_get_type_dependency_when_default_passed(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        context = alluka.Client().set_type_dependency(mock_type, mock_value).make_context()
        context = alluka.OverridingContext(context)

        assert context.get_type_dependency(mock_type, default=123) is mock_value

    def test_get_type_dependency_when_overriden(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        context = alluka.Client().set_type_dependency(mock_type, mock.Mock()).make_context()
        context = alluka.OverridingContext(context).set_type_dependency(mock_type, mock_value)

        assert context.get_type_dependency(mock_type) is mock_value

    def test_get_type_dependency_when_overriden_and_default_passed(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        context = alluka.Client().set_type_dependency(mock_type, mock.Mock()).make_context()
        context = alluka.OverridingContext(context).set_type_dependency(mock_type, mock_value)

        assert context.get_type_dependency(mock_type, default=123) is mock_value

    def test_get_type_dependency_when_overridden_only(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        context = alluka.Client().make_context()
        context = alluka.OverridingContext(context).set_type_dependency(mock_type, mock_value)

        assert context.get_type_dependency(mock_type) is mock_value

    def test_get_type_dependency_when_not_found(self):
        mock_type: typing.Any = mock.Mock()
        context = alluka.OverridingContext(alluka.Client().make_context())

        with pytest.raises(KeyError):
            context.get_type_dependency(mock_type)

    def test_get_type_dependency_when_not_found_with_default(self):
        mock_type: typing.Any = mock.Mock()
        context = alluka.OverridingContext(alluka.Client().make_context())

        value = context.get_type_dependency(mock_type, default=4444)

        assert value == 4444


class TestBasicContext:
    def test_get_type_dependency(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        mock_client = alluka.Client().set_type_dependency(mock_type, mock_value)
        ctx = alluka.BasicContext(mock_client)  # pyright: ignore[reportDeprecated]

        result = ctx.get_type_dependency(mock_type)

        assert result is mock_value

    def test_get_type_dependency_with_default(self):
        default = object()
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        mock_client = alluka.Client().set_type_dependency(mock_type, mock_value)
        ctx = alluka.BasicContext(mock_client)  # pyright: ignore[reportDeprecated]

        result = ctx.get_type_dependency(mock_type, default=default)

        assert result is mock_value

    def test_get_type_dependency_when_not_found(self):
        mock_type: typing.Any = mock.Mock()
        ctx = alluka.BasicContext(alluka.Client())  # pyright: ignore[reportDeprecated]

        with pytest.raises(KeyError):
            ctx.get_type_dependency(mock_type)

    def test_get_type_dependency_when_not_found_with_default(self):
        default = object()
        mock_type: typing.Any = mock.Mock()
        ctx = alluka.BasicContext(alluka.Client())  # pyright: ignore[reportDeprecated]

        result = ctx.get_type_dependency(mock_type, default=default)

        assert result is default

    def test_get_type_dependency_when_special_cased(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        mock_client = alluka.Client().set_type_dependency(mock_type, mock.Mock())
        ctx = alluka.BasicContext(mock_client)._set_type_special_case(  # pyright: ignore[reportDeprecated]
            mock_type, mock_value
        )

        result = ctx.get_type_dependency(mock_type)

        assert result is mock_value

    def test__remove_type_special_case(self):
        mock_type: typing.Any = mock.Mock()
        mock_value = mock.Mock()
        mock_client = alluka.Client().set_type_dependency(mock_type, mock_value)
        ctx = alluka.BasicContext(mock_client)._set_type_special_case(  # pyright: ignore[reportDeprecated]
            mock_type, mock.Mock()
        )

        ctx._remove_type_special_case(mock_type)  # pyright: ignore[reportDeprecated]

        assert ctx.get_type_dependency(mock_type) is mock_value

    def test__remove_type_special_case_when_not_set(self):
        mock_client = alluka.Client()
        mock_callback = mock.Mock()
        ctx = alluka.BasicContext(mock_client)  # pyright: ignore[reportDeprecated]

        with pytest.raises(KeyError):
            ctx._remove_type_special_case(mock_callback)  # pyright: ignore[reportDeprecated]



# -*- coding: utf-8 -*-
# Tanjun Examples - A collection of examples for Tanjun.
# Written in 2023 by Faster Speeding Lucina@lmbyrne.dev
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.

# pyright: reportUnusedFunction=none
# pyright: reportUnusedVariable=none

import typing

import alluka
import alluka.local


class Foo: ...


class BarResult: ...


bar_callback = BarResult


# fmt: off
def default_descriptor_example() -> None:
    def callback(
        foo: Foo = alluka.inject(type=Foo),
        bar: BarResult = alluka.inject(callback=bar_callback),
    ) -> None:
        ...
# fmt: on


def default_descriptor_inferred_type_example() -> None:
    async def callback(foo: Foo = alluka.inject()) -> None: ...


def type_hint_example() -> None:
    def callback(
        foo: typing.Annotated[Foo, alluka.inject(type=Foo)],
        bar: typing.Annotated[BarResult, alluka.inject(callback=bar_callback)],
    ) -> None: ...


def injected_example() -> None:
    async def callback(foo: alluka.Injected[Foo]) -> None: ...


# fmt: off
async def calling_with_di_example() -> None:
    client = alluka.Client()

    async def callback(
        argument: int,
        /,
        injected: alluka.Injected[Foo],
        keyword_arg: str,
    ) -> int:
        ...

    ...

    result = await client.call_with_async_di(callback, 123, keyword_arg="ok")
# fmt: on


# fmt: off
def calling_with_di_sync_example() -> None:
    client = alluka.Client()

    def callback(
        argument: int,
        /,
        injected: alluka.Injected[Foo],
        keyword_arg: str,
    ) -> int:
        ...

    ...

    result = client.call_with_di(callback, 123, keyword_arg="ok")
# fmt: on


def other_callback() -> None: ...


def context_di_example() -> None:
    def foo(ctx: alluka.Injected[alluka.abc.Context]) -> None:
        result = ctx.call_with_di(other_callback, 542, keyword_arg="meow")


def async_context_di_example() -> None:
    async def bar(ctx: alluka.Injected[alluka.abc.Context]) -> None:
        result = await ctx.call_with_async_di(other_callback, 123, keyword_arg="ok")


class TypeA: ...


TypeB = TypeA
type_a_impl = TypeA()
type_b_impl = TypeB()


# fmt: off
def setting_dependencies_example() -> None:
    client = (
        alluka.Client()
        .set_type_dependency(TypeA, type_a_impl)
        .set_type_dependency(TypeB, type_b_impl)
    )
# fmt: on


default_callback = other_callback


def override_callback_example() -> None:
    client = alluka.Client().set_callback_override(default_callback, other_callback)


async def initialize_example() -> None:
    client = alluka.local.initialize()
    client.set_type_dependency(TypeA, type_a_impl)

    ...

    async def callback(value: TypeA = alluka.inject()) -> None: ...

    result = await alluka.local.call_with_async_di(callback)


async def async_callback() -> None: ...


async def scoped_example() -> None:
    async def callback() -> None:
        result = await alluka.local.call_with_async_di(async_callback)

    with alluka.local.scope_client() as client:
        client.set_type_dependency(TypeA, type_a_impl)

        await callback()


async def local_auto_inject_example() -> None:
    @alluka.local.auto_inject_async
    async def callback(value: TypeA = alluka.inject()) -> None: ...

    with alluka.local.scope_client() as client:
        client.set_type_dependency(TypeA, type_a_impl)

        await callback()


def auto_inject_example() -> None:
    client = alluka.Client()

    @client.auto_inject
    def callback(other_arg: str, value: TypeA = alluka.inject()) -> None: ...

    callback(other_arg="beep")  # `value` will be injected.


async def auto_inject_async_example() -> None:
    client = alluka.Client()

    @client.auto_inject_async
    async def callback(value: TypeA = alluka.inject()) -> None: ...

    await callback()  # `value` will be injected.


def set_component_maker_example() -> None:
    client = alluka.Client().set_make_context(alluka.CachingContext)


def overriding_context_example() -> None:
    def callback(ctx: alluka.Injected[alluka.abc.Context]) -> None:
        ctx = alluka.OverridingContext(ctx).set_type_dependency(TypeA, type_a_impl)

        ctx.call_with_di(other_callback)


def overriding_context_from_client_example() -> None:
    client = alluka.Client().set_type_dependency(TypeA, type_a_impl)

    ctx = alluka.OverridingContext.from_client(client).set_type_dependency(TypeB, type_b_impl)

    ctx.call_with_di(other_callback)


def caching_example() -> None:
    client = alluka.Client().set_make_context(alluka.CachingContext)
    state = 0

    def injected_callback() -> int:
        nonlocal state
        state += 1

        return state

    def callback(
        result: int = alluka.inject(callback=injected_callback),
        other_result: int = alluka.inject(callback=injected_callback),
    ) -> None:
        print(result)
        print(other_result)

    client.call_with_di(callback)
    print("-")
    client.call_with_di(callback)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import typing

import mock
import pytest

import alluka

if typing.TYPE_CHECKING:
    from collections import abc as collections

    from typing_extensions import Self

    _CallbackT = typing.TypeVar("_CallbackT", bound=collections.Callable[..., typing.Any])
    _DefaultT = typing.TypeVar("_DefaultT")
    _T = typing.TypeVar("_T")
    _DefaultT = typing.TypeVar("_DefaultT")

    _CoroT = collections.Coroutine[typing.Any, typing.Any, _T]


class MockClient(alluka.abc.Client):
    def make_context(self) -> alluka.abc.Context:
        raise NotImplementedError

    def as_async_self_injecting(
        self, callback: _CallbackT, /
    ) -> alluka.abc.AsyncSelfInjecting[_CallbackT]:  # pyright: ignore[reportDeprecated]
        raise NotImplementedError

    def as_self_injecting(
        self, callback: _CallbackT, /
    ) -> alluka.abc.SelfInjecting[_CallbackT]:  # pyright: ignore[reportDeprecated]
        raise NotImplementedError

    @typing.overload
    def call_with_ctx(
        self,
        ctx: alluka.abc.Context,
        callback: collections.Callable[..., _CoroT[typing.Any]],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.NoReturn: ...

    @typing.overload
    def call_with_ctx(
        self, ctx: alluka.abc.Context, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T: ...

    def call_with_ctx(
        self, ctx: alluka.abc.Context, callback: collections.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T:
        raise NotImplementedError

    async def call_with_ctx_async(
        self, ctx: alluka.abc.Context, callback: alluka.abc.CallbackSig[_T], *args: typing.Any, **kwargs: typing.Any
    ) -> _T:
        raise NotImplementedError

    def set_type_dependency(self, type_: type[_T], value: _T, /) -> Self:
        raise NotImplementedError

    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT = ...) -> typing.Union[_T, _DefaultT]:
        raise NotImplementedError

    def remove_type_dependency(self, type_: type[typing.Any], /) -> Self:
        raise NotImplementedError

    def set_callback_override(
        self, callback: alluka.abc.CallbackSig[_T], override: alluka.abc.CallbackSig[_T], /
    ) -> Self:
        raise NotImplementedError

    def get_callback_override(
        self, callback: alluka.abc.CallbackSig[_T], /
    ) -> typing.Optional[alluka.abc.CallbackSig[_T]]:
        raise NotImplementedError

    def remove_callback_override(self, callback: alluka.abc.CallbackSig[_T], /) -> Self:
        raise NotImplementedError


class TestClient:
    def test_auto_inject(self):
        client = MockClient()
        client.call_with_di = mock.Mock()
        mock_callback = mock.Mock()

        wrapped = client.auto_inject(mock_callback)

        result = wrapped(341, "123", hello="bye")

        client.call_with_di.assert_called_once_with(mock_callback, 341, "123", hello="bye")
        assert result is client.call_with_di.return_value

    @pytest.mark.anyio
    async def test_auto_inject_async(self):
        client = MockClient()
        client.call_with_async_di = mock.AsyncMock()
        mock_callback = mock.AsyncMock()

        wrapped = client.auto_inject_async(mock_callback)

        result = await wrapped("555", 12312, kill="the fuckers")

        client.call_with_async_di.assert_awaited_once_with(mock_callback, "555", 12312, kill="the fuckers")
        assert result is client.call_with_async_di.return_value

    def test_call_with_di(self):
        client = MockClient()
        client.make_context = mock.Mock()
        mock_callback = mock.Mock()

        result = client.call_with_di(mock_callback, 333, "lll", meep="boop")

        assert result is client.make_context.return_value.call_with_di.return_value
        client.make_context.assert_called_once_with()
        client.make_context.return_value.call_with_di.assert_called_once_with(mock_callback, 333, "lll", meep="boop")

    @pytest.mark.anyio
    async def test_call_with_async_di(self):
        client = MockClient()
        client.make_context = mock.Mock()
        client.make_context.return_value.call_with_async_di = mock.AsyncMock()
        mock_callback = mock.AsyncMock()

        result = await client.call_with_async_di(mock_callback, "777", 555, meep="3222")

        assert result is client.make_context.return_value.call_with_async_di.return_value
        client.make_context.assert_called_once_with()
        client.make_context.return_value.call_with_async_di.assert_awaited_once_with(
            mock_callback, "777", 555, meep="3222"
        )


class MockContext(alluka.abc.Context):
    @property
    def injection_client(self) -> alluka.abc.Client:
        raise NotImplementedError

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /) -> _T: ...

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT = ...) -> typing.Union[_T, _DefaultT]:
        raise NotImplementedError


class TestContext:
    def test_cache_result(self):
        mock_callback = mock.Mock()
        ctx = MockContext()
        ctx.cache_result(mock_callback, mock.Mock())

    def test_get_cached_result(self):
        ctx = alluka.Context(alluka.Client())

        with pytest.raises(KeyError):
            ctx.get_cached_result(mock.Mock())

    def test_get_cached_result_when_defaulting(self):
        ctx = alluka.Context(alluka.Client())
        default = mock.Mock()

        value = ctx.get_cached_result(mock.Mock(), default=default)

        assert value is default



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Internal types used by Alluka."""

from __future__ import annotations

__all__: list[str] = ["Injected", "InjectedDescriptor"]

import enum
import typing

from . import _errors  # pyright: ignore[reportPrivateUsage]
from . import abc as alluka

if typing.TYPE_CHECKING:
    from collections import abc as collections


_T = typing.TypeVar("_T")


class _UndefinedEnum(enum.Enum):
    UNDEFINED = object()


UNDEFINED = _UndefinedEnum.UNDEFINED
"""Singleton used internally to indicate that a value is undefined."""
UndefinedOr = typing.Union[_T, typing.Literal[_UndefinedEnum.UNDEFINED]]
"""Union for a value which may be undefined."""


class InjectedCallback:
    """Descriptor of a callback that's being used to resolve a paremeter's value."""

    __slots__ = ("callback",)

    def __init__(self, callback: alluka.CallbackSig[typing.Any], /) -> None:
        """Initialize the callback descriptor.

        Parameters
        ----------
        callback
            The callback to use to resolve the parameter's value.
        """
        self.callback = callback

    def resolve(self, ctx: alluka.Context) -> typing.Any:
        """Synchronously resolve the callback.

        !!! warning
            Unlike [InjectedCallback.resolve_async][], this method will block the
            current thread and does not support async callback dependencies.

        Parameters
        ----------
        ctx
            The context to use when resolving the callback.

        Raises
        ------
        alluka.SyncOnlyError
            If the callback or any of its callback dependencies are async.
        alluka.MissingDependencyError
            If any of the callback's type dependencies aren't implemented by
            the context's client.
        """
        callback = ctx.injection_client.get_callback_override(self.callback) or self.callback
        cached = ctx.get_cached_result(callback, default=UNDEFINED)

        if cached is UNDEFINED:
            result = ctx.injection_client.call_with_ctx(ctx, callback)
            ctx.cache_result(callback, result)
            return result

        return cached

    def resolve_async(self, ctx: alluka.Context) -> collections.Coroutine[typing.Any, typing.Any, typing.Any]:
        """Asynchronously resolve the callback.

        Parameters
        ----------
        ctx
            The context to use when resolving the callback.

        Raises
        ------
        alluka.MissingDependencyError
            If any of the callback's type dependencies aren't implemented by
            the context's client.
        """
        callback = ctx.injection_client.get_callback_override(self.callback) or self.callback
        cached = ctx.get_cached_result(callback, default=UNDEFINED)

        if cached is UNDEFINED:
            result = ctx.injection_client.call_with_ctx_async(ctx, callback)
            ctx.cache_result(callback, result)
            return result

        return cached


class InjectedType:
    """Descriptor of a type that a parameter's value is being resolved to."""

    __slots__ = ("default", "repr_type", "types")

    def __init__(
        self,
        repr_type: typing.Any,
        types: collections.Sequence[type[typing.Any]],
        /,
        *,
        default: UndefinedOr[typing.Any] = UNDEFINED,
    ) -> None:
        """Initialize the type descriptor.

        Parameters
        ----------
        base_type
            The type to resolve to.
        default
            The default value to use if the type can't be resolved.

            Without a default, any attempts to resolve a type that isn't implemented
            by the linked client will lead to [alluka.MissingDependencyError][].
        """
        self.default = default
        self.repr_type = repr_type
        self.types = types

    def resolve(self, ctx: alluka.Context) -> typing.Any:
        """Resolve the type.

        Parameters
        ----------
        ctx
            The context to use when resolving the type.

        Returns
        -------
        typing.Any
            The resolved type.
        """
        for cls in self.types:
            if (result := ctx.get_type_dependency(cls, default=UNDEFINED)) is not UNDEFINED:
                return result

        if self.default is not UNDEFINED:
            return self.default

        raise _errors.MissingDependencyError(
            f"Couldn't resolve injected type(s) {self.repr_type} to actual value", self.repr_type
        ) from None


class InjectedTypes(int, enum.Enum):
    """Enum of the different types of injected values."""

    CALLBACK = enum.auto()
    """An injeted callback.

    The result of the callback will be used as the value of the parameter.
    """

    TYPE = enum.auto()
    """An injected type.

    An implementation of a this type from the linked client will be injected
    as the value of the parameter.
    """


InjectedTuple = typing.Union[
    tuple[typing.Literal[InjectedTypes.CALLBACK], InjectedCallback],
    tuple[typing.Literal[InjectedTypes.TYPE], InjectedType],
]
"""Type of the tuple used to describe an injected value."""

_TypeT = type[_T]


class InjectedDescriptor(typing.Generic[_T]):
    """Descriptor used to a declare keyword-argument as requiring an injected dependency.

    This is the type returned by [alluka.inject][].
    """

    __slots__ = ("callback", "type")

    callback: typing.Optional[alluka.CallbackSig[_T]]
    """The callback to use to resolve the parameter's value.

    If this is `None` then this is a type dependency.
    """

    type: typing.Optional[_TypeT[_T]]  # noqa: VNE003
    """The type to use to resolve the parameter's value.

    If both this and `callback` are `None`, then this is a type dependency
    and the type will be inferred from the parameter's annotation.
    """

    def __init__(
        self,
        *,
        callback: typing.Optional[alluka.CallbackSig[_T]] = None,
        type: typing.Optional[_TypeT[_T]] = None,  # noqa: A002
    ) -> None:  # TODO: add default/factory to this?
        """Initialise an injection default descriptor.

        !!! note
            If neither `type` or `callback` is provided, an injected type
            will be inferred from the argument's annotation.

        Parameters
        ----------
        callback
            The callback to use to resolve the dependency.

            If this callback has no type dependencies then this will still work
            without an injection context but this can be overridden using
            [alluka.abc.Client.set_callback_override][].
        type
            The type of the dependency to resolve.

            If a union (e.g. `typing.Union[A, B]`, `A | B`, `typing.Optional[A]`)
            is passed for `type` then each type in the union will be tried
            separately after the litarl union type is tried, allowing for resolving
            `A | B` to the value set by `set_type_dependency(B, ...)`.

            If a union has `None` as one of its types (including `Optional[T]`)
            then `None` will be passed for the parameter if none of the types could
            be resolved using the linked client.

        Raises
        ------
        ValueError
            If both `callback` and `type` are provided.
        """
        if callback is not None and type is not None:
            raise ValueError("Only one of `callback` or `type` can be specified")

        self.callback = callback
        self.type = type

    def __getattr__(self, __name: str) -> typing.NoReturn:
        raise AttributeError(
            "Tried accessing a parameter that was not injected yet. Did you forget to inject dependencies?"
        )


Injected = typing.Annotated[_T, InjectedTypes.TYPE]
"""Type alias used to declare a keyword argument as requiring an injected type.

If a union (e.g. `typing.Union[A, B]`, `A | B`, `typing.Optional[A]`)
is passed then each type in the union will be tried separately rather than
the literal type, allowing for resolving `A | B` to the value set by
`set_type_dependency(B, ...)`.

If a union has `None` as one of its types (including `Optional[T]`)
then `None` will be passed for the parameter if none of the types could
be resolved using the linked client.

!!! note
    This is a [typing.Annotated][] alias and the behaviour for nested
    Annotated types may be found at the docs for it [typing.Annotated][].
"""



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Alluka's standard injection context implementations."""

# pyright: reportOverlappingOverload=warning

from __future__ import annotations

__all__: list[str] = ["BasicContext", "CachingContext", "Context", "OverridingContext"]

import enum
import typing

import typing_extensions

from . import abc as alluka

if typing.TYPE_CHECKING:
    from typing_extensions import Self


_T = typing.TypeVar("_T")


_DefaultT = typing.TypeVar("_DefaultT")


class _NoDefaultEnum(enum.Enum):
    VALUE = object()


_NO_VALUE: typing.Literal[_NoDefaultEnum.VALUE] = _NoDefaultEnum.VALUE
_NoValueOr = typing.Union[_T, typing.Literal[_NoDefaultEnum.VALUE]]


class Context(alluka.Context):
    """Basic implementation of [alluka.abc.Context][] without caching."""

    __slots__ = ("_client",)

    def __init__(self, client: alluka.Client, /) -> None:
        """Initialise an injection context.

        Parameters
        ----------
        client
            The injection client this context is bound to.
        """
        self._client = client

    @property
    def injection_client(self) -> alluka.Client:
        # <<inherited docstring from alluka.abc.Context>>.
        return self._client

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /) -> _T: ...

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    def get_type_dependency(
        self, type_: type[_T], /, *, default: _NoValueOr[_DefaultT] = _NO_VALUE
    ) -> typing.Union[_T, _DefaultT]:
        # <<inherited docstring from alluka.abc.Context>>.
        if type_ is alluka.Context:
            return self  # type: ignore

        result = self._client.get_type_dependency(type_, default=default)

        if result is _NO_VALUE:
            raise KeyError

        return result


class CachingContext(Context):
    """Basic implementation of [alluka.abc.Context][] with callback result caching."""

    __slots__ = ("_result_cache",)

    def __init__(self, client: alluka.Client, /) -> None:
        """Initialise a caching injection context.

        Parameters
        ----------
        client
            The injection client this context is bound to.
        """
        super().__init__(client)
        self._result_cache: dict[alluka.CallbackSig[typing.Any], typing.Any] = {}

    def cache_result(self, callback: alluka.CallbackSig[_T], value: _T, /) -> None:
        # <<inherited docstring from alluka.abc.Context>>.
        self._result_cache[callback] = value

    @typing.overload
    def get_cached_result(self, callback: alluka.CallbackSig[_T], /) -> _T: ...

    @typing.overload
    def get_cached_result(
        self, callback: alluka.CallbackSig[_T], /, *, default: _DefaultT
    ) -> typing.Union[_T, _DefaultT]: ...

    def get_cached_result(
        self, callback: alluka.CallbackSig[_T], /, *, default: _NoValueOr[_DefaultT] = _NO_VALUE
    ) -> typing.Union[_T, _DefaultT]:
        # <<inherited docstring from alluka.abc.Context>>.
        result = self._result_cache.get(callback, default)

        if result is _NO_VALUE:
            raise KeyError

        return result


class OverridingContext(alluka.Context):
    """Context which exists to override an existing context with special-cased type-dependencies.

    Examples
    --------
    ```py
    def other_callback(value: alluka.Inject[Type]) -> None:
        ...

    def callback(ctx: alluka.Inject[alluka.abc.Context]) -> None:
        ctx = alluka.OverridingContext(ctx).set_type_dependency(Type, value)

        ctx.call_with_di(other_callback)
    ```

    ```py
    client = alluka.abc.Client().set_type_dependency(Type, value)
    ctx = alluka.OverridingContext.from_client(client).set_type_dependency(OtherType, value)
    ```
    """

    __slots__ = ("_context", "_overrides")

    def __init__(self, context: alluka.Context, /) -> None:
        """Initialise an overriding context.

        While this is designed to wrap an existing context,
        [OverridingContext.from_client][alluka.OverridingContext.from_client] can
        be used to create this from an alluka client.

        Parameters
        ----------
        context
            The context to wrap.
        """
        self._context = context
        self._overrides: dict[type[typing.Any], typing.Any] = {}

    @classmethod
    def from_client(cls, client: alluka.Client, /) -> Self:
        """Create an overriding context from an injection client.

        This will wrap the context returned by [Client.make_context][alluka.abc.Client.make_context].

        Parameters
        ----------
        client
            The alluka client to make an overriding context for.

        Returns
        -------
        OverridingContext
            The created overriding context.
        """
        return cls(client.make_context())

    @property
    def injection_client(self) -> alluka.Client:
        # <<inherited docstring from alluka.abc.Context>>.
        return self._context.injection_client

    @typing.overload
    def get_cached_result(self, callback: alluka.CallbackSig[_T], /) -> _T: ...

    @typing.overload
    def get_cached_result(
        self, callback: alluka.CallbackSig[_T], /, *, default: _DefaultT
    ) -> typing.Union[_T, _DefaultT]: ...

    def get_cached_result(
        self, callback: alluka.CallbackSig[_T], /, *, default: _NoValueOr[_DefaultT] = _NO_VALUE
    ) -> typing.Union[_T, _DefaultT]:
        value = self._context.get_cached_result(callback, default=default)

        if value is _NO_VALUE:
            raise KeyError

        return value

    def cache_result(self, callback: alluka.CallbackSig[_T], value: _T, /) -> None:
        self._context.cache_result(callback, value)

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /) -> _T: ...

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    def get_type_dependency(
        self, type_: type[_T], /, *, default: _NoValueOr[_DefaultT] = _NO_VALUE
    ) -> typing.Union[_T, _DefaultT]:
        # <<inherited docstring from alluka.abc.Context>>.
        value = self._overrides.get(type_, default)
        if value is default:
            value = self._context.get_type_dependency(type_, default=default)

        if value is _NO_VALUE:
            raise KeyError

        return value

    def set_type_dependency(self, type_: type[_T], value: _T, /) -> Self:
        """Add a context specific type dependency.

        Parameters
        ----------
        type_
            The type of the dependency to add an implementation for.
        value
            The value of the dependency.

        Returns
        -------
        Self
            The context to allow chaining.
        """
        self._overrides[type_] = value
        return self


@typing_extensions.deprecated("Use CachingContext")
class BasicContext(CachingContext):
    """Deprecated alias of [alluka.CachingContext][].

    [alluka.OverridingContext][] replaces the undocumented type special casing.
    """

    __slots__ = ("_special_case_types",)

    def __init__(self, client: alluka.Client, /) -> None:
        super().__init__(client)
        self._special_case_types: dict[type[typing.Any], typing.Any] = {}

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /) -> _T: ...

    @typing.overload
    def get_type_dependency(self, type_: type[_T], /, *, default: _DefaultT) -> typing.Union[_T, _DefaultT]: ...

    def get_type_dependency(
        self, type_: type[_T], /, *, default: _NoValueOr[_DefaultT] = _NO_VALUE
    ) -> typing.Union[_T, _DefaultT]:
        # <<inherited docstring from alluka.abc.Context>>.
        value = self._special_case_types.get(type_, default)
        if value is default:
            value = super().get_type_dependency(type_, default=default)

        if value is _NO_VALUE:
            raise KeyError

        return value

    @typing_extensions.deprecated("Use ContextOverride")
    def _set_type_special_case(self, type_: type[_T], value: _T, /) -> Self:
        if not self._special_case_types:
            self._special_case_types = {}

        self._special_case_types[type_] = value
        return self

    @typing_extensions.deprecated("Use ContextOverride")
    def _remove_type_special_case(self, type_: type[typing.Any], /) -> Self:
        if not self._special_case_types:
            raise KeyError(type_)

        del self._special_case_types[type_]
        return self



import alluka

client = alluka.Client().set_make_context(alluka.CachingContext)
state = 0

def injected_callback() -> int:
    global state
    state += 1

    return state

def callback(
    result: int = alluka.inject(callback=injected_callback),
    other_result: int = alluka.inject(callback=injected_callback),
) -> None:
    print(result)
    print(other_result)


client.call_with_di(callback)
